/* mexrdme.c */

/* J. Cullhed 2008-08-04. */

#include <stdlib.h>
#include <mex.h>
#include <matrix.h>
#include "report.h"
#include "propensities.h"
#include "rdme.h"

/* to allow for -largeArrayDims */
#ifndef mwIndex
#define mwIndex int
#endif
#ifndef mwSize
#define mwSize int
#endif

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
  /* Check syntax. */
  if (nrhs!=12 || nlhs!=1)
    mexErrMsgTxt("Wrong number of arguments.");

  /* Load arguments. */
  const mxArray *mxTspan=prhs[0];
  const mxArray *mxU0=prhs[1];
  const mxArray *mxD=prhs[2];
  const mxArray *mxN=prhs[3];
  const mxArray *mxG=prhs[4];
  const mxArray *mxK=prhs[5];
  const mxArray *mxI=prhs[6];
  const mxArray *mxS=prhs[7];
  const mxArray *mxVol=prhs[8];
  const mxArray *mxData=prhs[9];
  const mxArray *mxSd=prhs[10];
  const mxArray *mxOPTS=prhs[11];

  /* Get dimensions. */
  const int tlen=mxGetNumberOfElements(mxTspan);
  const size_t Ncells=mxGetNumberOfElements(mxVol);
  const int Mspecies=mxGetM(mxN);
  const int Mreactions=mxGetN(mxN);
  const size_t M1=mxGetN(mxK);
  const int dsize=mxGetM(mxData);
  const size_t Ndofs=Ncells*Mspecies;

  /* Get pointers to non sparse objects. */
  const double *tspan=mxGetPr(mxTspan);
  const double *u0_double=mxGetPr(mxU0);
  const double *vol=mxGetPr(mxVol);
  const double *data=mxGetPr(mxData);
  const double *sd_double=mxGetPr(mxSd);
  const double *OPTS=mxGetPr(mxOPTS);

  /* Get sparse matrix D. */
  const mwIndex *jcD=(const mwIndex *)mxGetJc(mxD);
  const mwIndex *irD=(const mwIndex *)mxGetIr(mxD);
  const double *prD=mxGetPr(mxD);

  /* Get sparse matrix N. */
  const mwIndex *jcN=(const mwIndex *)mxGetJc(mxN);
  const mwIndex *irN=(const mwIndex *)mxGetIr(mxN);
  const double *prN_double=mxGetPr(mxN);
  const mwIndex nnzN=jcN[mxGetN(mxN)];

  /* Get sparse matrix G. */
  const mwIndex *jcG=(const mwIndex *)mxGetJc(mxG);
  const mwIndex *irG=(const mwIndex *)mxGetIr(mxG);

  /* Get pointers to K and I. */
  const double *K=mxGetPr(mxK);
  const double *I_double=mxGetPr(mxI);

  /* Get sparse matrix S. */
  const mwIndex *jcS=(const mwIndex *)mxGetJc(mxS);
  const double *prS_double=mxGetPr(mxS);
  const size_t nnzS=jcS[mxGetN(mxS)];
  
  /* Typecast from double to int. */
  int *u0=mxMalloc(Ndofs*sizeof(int));
  int *sd=mxMalloc(Ncells*sizeof(int));
  int *prN=mxMalloc(nnzN*sizeof(int));
  int *I=mxMalloc(3*M1*sizeof(int));
  int *prS=mxMalloc(nnzS*sizeof(int));

  size_t i;
  for (i=0; i<Ndofs; i++) u0[i]=(int)u0_double[i];
  for (i=0; i<Ncells; i++) sd[i]=(int)sd_double[i];
  for (i=0; i<nnzN; i++) prN[i]=(int)prN_double[i];
  for (i=0; i<3*M1; i++) I[i]=(int)I_double[i]-1;
  for (i=0; i<nnzS; i++) prS[i]=(int)prS_double[i];

  /* Set up result matrix U. */
  int *U;
  U=mxMalloc(Ndofs*tlen*sizeof(int));

  /* Parse options. */
  ReportFun report=NULL;
  int report_level=0;
  if ((int)OPTS[0]) {
    report=&reportFun1;
    report_level=(int)OPTS[0];
  }
  if (!mxIsNaN(OPTS[1])) srand48((long)OPTS[1]);

  /* Allocate the propensity list (1st requirement of the model file). */
  PropensityFun *propList = Mreactions == M1 ? 0 : ALLOC_propensities();

  /* Master call. */
  rdme_solve((const unsigned *)irD, (const unsigned *)jcD, prD, u0,
      (const unsigned  *)irN, (const unsigned  *)jcN, prN, (const unsigned  *)irG,
      (const unsigned  *)jcG, K, I, (const unsigned  *)jcS, prS, tspan, tlen,
      propList, U, vol, data, sd, Ncells, Mspecies, Mreactions, M1, dsize,
      report, report_level);

  /* Deallocate propensities (2nd requirement of the model file). */
  if (Mreactions != M1) FREE_propensities(propList);

  /* Deallocate. */
  mxFree(prS);
  mxFree(I); 
  mxFree(prN);
  mxFree(sd);
  mxFree(u0);

  /* Put result in plhs[0] and typecast from int to double. */
  plhs[0]=mxCreateDoubleMatrix(Ndofs, tlen, mxREAL);
  double *U_out=mxGetPr(plhs[0]);
  for (i=0; i<Ndofs*tlen; i++) U_out[i]=(double)U[i];
  
  /* Deallocate. */
  mxFree(U);
}

