/* rdme.c */

/* J. Cullhed 2008-08-01. */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "propensities.h"
#include "rdme.h"
#include "binheap.h"

#if !defined(MALLOC) || !defined(FREE)
  #error "Must define MALLOC and FREE."
#endif

void *MALLOC(size_t);
void FREE(void *);

double l_inlineProp(const int *xx, const double *k, const int *i,
    const int *prS, size_t nS, double vol, int sd)
/* Inline propensity. Given constants k[0..2], indices i[0..2] and volume vol,
   this function computes an elementary propensity at state xx. */
{
  double res=0.0;

  /* Check if this propensity is turned off. */
  if (sd==0) return 0.0;
  for (size_t j=0; j<nS; j++) if (sd==prS[j]) return 0.0;

  /* Calculate inline propensity. */
  if (k[0]!=0.0) {
    if (i[0]!=i[1]) res+=k[0]*xx[i[0]]*xx[i[1]]/vol;
    else res+=k[0]*xx[i[0]]*(xx[i[0]]-1)/(2.0*vol);
  }
  if (k[1]!=0.0) res+=k[1]*xx[i[2]];

  res+=k[2]*vol;
  return res;
}

void rdme_solve(const unsigned *irD, const unsigned *jcD, const double *prD,
    const int *u0, const unsigned *irN, const unsigned *jcN, const int *prN,
    const unsigned *irG, const unsigned *jcG, const double *K, const int *I,
    const unsigned *jcS, const int *prS,
    const double *tspan, size_t tlen, PropensityFun *rfun, int *U,
    const double *vol, const double *data, const int *sd, size_t Ncells,
    size_t Mspecies, size_t Mreactions, size_t M1, size_t dsize,
    ReportFun report, int report_level)
/*

Input (not changed):

Ncells.
  Number of subvolumes.

Mspecies.
  Number of different species.

Mreactions.
  Total number of reactions.

M1.
  Number of inline propensities.

dsize.
  Size of data vector sent to propensities.

Defines:
  Ndofs=Ncells*Mspecies.

Diffusion matrix D. Double sparse (Ndofs X Ndofs).
  Macroscopic diffusion matrix, D(i,j) is the diffusion rate from dof #j to 
  dof #i. This matrix uses the CSR-format and not CSC because fast access to
  rows is needed.

Initial state vector u0. Integer (Mspecies X Ncells).
  Gives the initial number of species in each subvolume.

Stochiometric matrix N. Integer sparse (Mspecies X Nreactions).
  N(:,j) describes how reaction j changes the number of species.

Dependency graph G. Integer sparse (Mreactions X Mspecies+Mreactions).
  G(i,Mspecies+j) is non-zero if executing reaction j means that reaction i
  needs to be re-evaluated. The first Mspecies columns of G similarily cover
  diffusion events.

Inline propensity coefficients K. Double (3 X M1)
  K(:,i) gives the coefficients to reaction #i. Where i is in [0,M1).

Inline propensity index I. Integer (3 X M1)
  I(:,i) gives the indices used by reaction #i. Where i is in [0,M1].

Inline propensity subdomain S. Integer sparse (<=Ncells X M1)
  S(:,i) lists all subdomains in which the corresponding inline propensity
  is off.

tspan. Double vector.
  Output times. tspan[0] is the start time and tspan[length(tspan)-1] is the 
  stop time.

Propensity prop. Vector of function pointers (length Mreactions).
  The type of this vector is PropensityFun which defines the input to a property
  function.
  typedef double (*PropensityFun)(const int *xx, double vol, const double
    *data, int sd);

ReportFun report. Pointer to report function.
  This function will be called each time the simulation has reached a time
  index in tspan. The purpose is to show the progress to the user without
  dependency of any external interfaces. A NULL pointer means that no report
  should be made. This function is also called when the error flag errcode is
  nonzero.
  typedef void (*ReportFun)(double time, const double t0, const double tend,
    int *count, int errcode, int report_level);

Details:
  G, N and S are sparse matrices in compressed column format (CCS). D is sparse
  but in compressed row format (CRS) or equivalent a transposed matrix in
  CCS format.
  jcD, irD, prD (double *)
  jcN, irN, prN (int *)
  jcG, irG (int *)
  jcS, prS (int *)

Other inputs not covered above:
Ordering of the dofs: Dof #i is located in cell #(i/Mspecies),
and the dofs located in cell #j is u0(:,j). Thus, u0 is understood as a matrix
of size Mspecies X Ncells.

(cell -> vol) vol. Double vector (length Ncells).
  vol[i] gives the volume of cell #i.

(cell -> data) data. Double matrix (dsize X Ncells).
  Generalized data matrix, data(:,j) gives a data vector for cell #j.

(cell -> sd) sd. Integer vector (length Ncells).
  Subdomain number. sd[i] is the subdomain of cell #i. The vector sd can also
  be used to separate boundaries, line segments and points.

Output:
Report matrix U. (Ndofs X length(tspan)).
  U(:,j) contains the state of the system at tspan(j).

*/
{
  double told, tt=tspan[0];
  double rdelta, rrdelta;
  double rand, cum;
  double *srrate, *rrate;
  double *sdrate, *Ddiag;
  double *rtimes;
  double old_rrate=0.0, old_drate=0.0;
  int *node, *heap;
  int *xx;
  int count[4]={0,0,0,0};
  int subvol, event, re, spec, to_node, to_vol=0, errcode=0;
  size_t i, j, it=0;
  const size_t Ndofs=Ncells*Mspecies;

  /* Set xx to the initial state. */
  xx=(int *)MALLOC(Ndofs*sizeof(int));
  memcpy(xx, u0, Ndofs*sizeof(int));

  /* Create reaction rate matrix (Mreactions X Ncells) and total rate
     vector. */
  rrate=(double *)MALLOC(Mreactions*Ncells*sizeof(double));
  srrate=(double *)MALLOC(Ncells*sizeof(double));

  /* Calculate the propensity for every reaction and every subvolume. Also store
     the sum of the reaction intensities in each subvolume in srrate. */
  for (i=0; i<Ncells; i++) {
    srrate[i]=0.0;
    for (j=0; j<M1; j++) {
      rrate[i*Mreactions+j]= 
        l_inlineProp(&xx[i*Mspecies], &K[j*3], &I[j*3], &prS[jcS[j]],
            jcS[j+1]-jcS[j], vol[i], sd[i]);
      srrate[i]+=rrate[i*Mreactions+j];
    }
    for (; j<Mreactions; j++) {
      rrate[i*Mreactions+j]=
        (*rfun[j-M1])(&xx[i*Mspecies], vol[i], &data[i*dsize], sd[i]);
      srrate[i]+=rrate[i*Mreactions+j];
    }
  }

  /* Create total diffusion rate vector (length Mcells). */
  sdrate=(double *)MALLOC(Ncells*sizeof(double));

  /* Store the negative of D's diagonal in Ddiag. */
  Ddiag=(double *)MALLOC(Ndofs*sizeof(double));
  for (i=0; i<Ndofs; i++) {
    Ddiag[i]=0.0;
    for (j=jcD[i]; j<jcD[i+1]; j++)
      if (irD[j]==i) Ddiag[i]=-prD[j];
  }

  /* Calculate the total diffusion rate for each subvolume. */
  for(i=0; i<Ncells; i++) {
    sdrate[i]=0.0;
    for(j=0; j<Mspecies; j++)
      sdrate[i]+=Ddiag[i*Mspecies+j]*xx[i*Mspecies+j];
  }

  /* Create binary heap. */
  rtimes=(double *)MALLOC(Ncells*sizeof(double));
  node=(int *)MALLOC(Ncells*sizeof(int));
  heap=(int *)MALLOC(Ncells*sizeof(int));

  /* Insert times to next event. */
  for (i=0; i<Ncells; i++) {
    rtimes[i]=-log(1.0-drand48())/(srrate[i]+sdrate[i])+tspan[0];
    node[i]=i;
    heap[i]=i;
  }
  initialize_heap(rtimes, node, heap, Ncells);

  /* Main loop. */
  for (;;) {
    /* Get the subvolume in which the next event occurred. */
    told=tt;
    tt=rtimes[0];
    subvol=node[0];

    /* Record values and exit as needed. */
    if (told>=tspan[it]||isinf(tt)) {
      if (report) report(tspan[it], tspan[0], tspan[tlen-1], count, 0,
        report_level);
	  for (; it<tlen && (told>=tspan[it]||isinf(tt)); it++)
        memcpy(&U[Ndofs*it], &xx[0], Ndofs*sizeof(int));
      if (it>=tlen) break;
      
    }

    /* Get random number. */
    rand=drand48();

    /* Reaction or Diffusion event? */
    if (rand*(srrate[subvol]+sdrate[subvol]) < srrate[subvol]) {
      /* Reaction event. */
      event=0;

      /* a) Determine the reaction re that occur. */
      rand*=(srrate[subvol]+sdrate[subvol]);
      for (re=0, cum=rrate[subvol*Mreactions]; rand>cum;
          re++, cum+=rrate[subvol*Mreactions+re]);

      /* b) Update the state of the subvolume 'subvol' and sdrate[subvol]. */
      for (i=jcN[re]; i<jcN[re+1]; i++) {
        xx[subvol*Mspecies+irN[i]]-=prN[i];
        if (xx[subvol*Mspecies+irN[i]]<0) errcode=1;
        sdrate[subvol]-=Ddiag[subvol*Mspecies+irN[i]]*prN[i];
      }

      /* c) Recalculate srrate[subvol] using dependency graph. */
      for (i=jcG[Mspecies+re], rdelta=0.0; i<jcG[Mspecies+re+1]; i++) {
        const double old=rrate[subvol*Mreactions+irG[i]];
        j=irG[i];

        if (j<M1)
          rdelta+=(rrate[subvol*Mreactions+j]=l_inlineProp(&xx[subvol*
              Mspecies], &K[j*3], &I[j*3], &prS[jcS[j]], jcS[j+1]-jcS[j],
              vol[subvol], sd[subvol]))-old;
        else
          rdelta+=(rrate[subvol*Mreactions+j]=(*rfun[j-M1])(&xx[subvol*
            Mspecies], vol[subvol], &data[subvol*dsize], sd[subvol]))-old;
      }
      srrate[subvol]+=rdelta;
    } else {
      /* Diffusion event. */
      event=1;

      /* a) Determine the species that diffuses. */
      rand*=(sdrate[subvol]+srrate[subvol]);
      rand-=srrate[subvol];
      for (spec=0, cum=Ddiag[subvol*Mspecies]*xx[subvol*Mspecies]; rand>cum;
          spec++, cum+=Ddiag[subvol*Mspecies+spec]*xx[subvol*Mspecies+spec]);

      /* b) Determine the direction of the diffusion. */
      const size_t col=subvol*Mspecies+spec;
      rand=drand48()*Ddiag[col];

      /* Search for diffusion direction. */
      for (i=jcD[col], cum=0.0; i<jcD[col+1]; i++)
        if (irD[i]!=col && (cum+=prD[i])>rand)
           break;
      /* paranoia fix: */
      if (i>=jcD[col+1]) i--;

      to_node=irD[i];
      to_vol=to_node/Mspecies;

      /* c) Execute the diffusion event (check for negative elements). */
      xx[subvol*Mspecies+spec]--;
      if (xx[subvol*Mspecies+spec]<0) errcode=1;
      xx[to_node]++;

      /* Save reaction and diffusion rates. */
      old_rrate=srrate[to_vol];
      old_drate=sdrate[to_vol];

      /* d) Recalculate the reaction rates using dependency graph G. */
      for (i=jcG[spec], rdelta=0.0, rrdelta=0.0; i<jcG[spec+1]; i++) {
        double old=rrate[subvol*Mreactions+irG[i]];
        j=irG[i];

        if (j<M1) {
          rdelta+=(rrate[subvol*Mreactions+j]=l_inlineProp(&xx[subvol*
              Mspecies], &K[j*3], &I[j*3], &prS[jcS[j]], jcS[j+1]-jcS[j],
              vol[subvol], sd[subvol]))-old;
 
          old=rrate[to_vol*Mreactions+j];
          rrdelta+=(rrate[to_vol*Mreactions+j]=l_inlineProp(&xx[to_vol*
              Mspecies], &K[j*3], &I[j*3], &prS[jcS[j]], jcS[j+1]-jcS[j],
              vol[to_vol], sd[to_vol]))-old;
        } else {
          rdelta+=(rrate[subvol*Mreactions+j]=(*rfun[j-M1])(&xx[subvol*
              Mspecies], vol[subvol], &data[subvol*dsize], sd[subvol]))-old;

          old=rrate[to_vol*Mreactions+j];
          rrdelta+=(rrate[to_vol*Mreactions+j]=(*rfun[j-M1])(&xx[to_vol*
              Mspecies], vol[to_vol], &data[to_vol*dsize], sd[to_vol]))-old;
        }
      }
      srrate[subvol]+=rdelta;
      srrate[to_vol]+=rrdelta;

      /* Adjust diffusion rates. */
      sdrate[subvol]-=Ddiag[subvol*Mspecies+spec];
      sdrate[to_vol]+=Ddiag[to_vol*Mspecies+spec];
    }

    /* Compute time to new event for this subvolume. */
    if(srrate[subvol]+sdrate[subvol]>0.0)
      rtimes[0]=-log(1.0-drand48())/(srrate[subvol]+sdrate[subvol])+tt;
    else
      rtimes[0]=INFINITY;
    
    /* Update the heap. */
    update(0, rtimes, node, heap, Ncells);

    /* If it was a diffusion event, also update the other affected
       node. */
    if(event) {
      if(srrate[to_vol]+sdrate[to_vol]>0.0) {
        if (!isinf(rtimes[heap[to_vol]]))
          rtimes[heap[to_vol]]=(old_rrate+old_drate)/(srrate[to_vol]+
              sdrate[to_vol])*(rtimes[heap[to_vol]]-tt)+tt;
        else
          rtimes[heap[to_vol]]=
              -log(1.0-drand48())/(srrate[to_vol]+sdrate[to_vol])+tt;
      } else
        rtimes[heap[to_vol]]=INFINITY;

      update(heap[to_vol], rtimes, node, heap, Ncells);

      /* Add diffusion event. */
      if (++count[3]==1000000000) {
        count[3]=0;
        count[2]++;
      }
    } else {
      /* Add reaction event. */
      if (++count[1]==1000000000) {
        count[1]=0;
        count[0]++;
      }
    }

    /* Check for error codes. */
    if (errcode) {
      /* Report the error that occurred. */
      if (report) report(tt, tspan[0], tspan[tlen-1], count, errcode,
        report_level);

      /* Cannot continue. Clear this solution and exit. */
      memcpy(&U[Ndofs*it], &xx[0], Ndofs*sizeof(int));
      memset(&U[Ndofs*(it+1)], 0, Ndofs*(tlen-it-1)*sizeof(int));
      break;
    }
  }

  /* Free memory. */
  FREE(heap);
  FREE(node);
  FREE(rtimes);
  FREE(Ddiag);
  FREE(sdrate);
  FREE(srrate);
  FREE(rrate);
  FREE(xx);
}

